import subprocess
import os

from bzrlib.workingtree import WorkingTree

from .utils import ErrorExit


class Vcs(object):

    err_update = (
        "Could not update branch %(path)s from %(branch_url)s\n\n %(output)s")
    err_branch = "Could not branch %(branch_url)s to %(path)s\n\n %(output)s"
    err_is_mod = "Couldn't determine if %(path)s was modified\n\n %(output)s"
    err_pull = (
        "Could not pull branch @ %(branch_url)s to %(path)s\n\n %(output)s")
    err_cur_rev = (
        "Could not determine current revision %(path)s\n\n %(output)s")

    def __init__(self, path, origin, log):
        self.path = path
        self.log = log
        self.origin = origin

    def _call(self, args, error_msg, cwd=None, stderr=()):
        try:
            if stderr is not None and not stderr:
                stderr = subprocess.STDOUT
            output = subprocess.check_output(
                args, cwd=cwd or self.path, stderr=stderr)
        except subprocess.CalledProcessError, e:
            #print "vcs err", " ".join(args), "[dir: %s]" % cwd, e.output
            self.log.error(error_msg % self.get_err_msg_ctx(e))
            raise ErrorExit()
        return output.strip()

    def get_err_msg_ctx(self, e):
        return {
            'path': self.path,
            'branch_url': self.origin,
            'exit_code': e.returncode,
            'output': e.output,
            'vcs': self.__class__.__name__.lower()}

    def get_cur_rev(self):
        raise NotImplementedError()

    def update(self, rev=None):
        raise NotImplementedError()

    def branch(self):
        raise NotImplementedError()

    def pull(self):
        raise NotImplementedError()

    def is_modified(self):
        raise NotImplementedError()

    # upstream missing revisions?


class Bzr(Vcs):

    def get_cur_rev(self):
        params = ["bzr", "revno", "--tree"]
        return self._call(params, self.err_cur_rev, stderr=None)

    def update(self, rev=None):
        params = ["bzr", "up"]
        if rev:
            params.extend(["-r", str(rev)])
        self._call(params, self.err_update)

    def branch(self):
        params = ["bzr", "co", "--lightweight", self.origin, self.path]
        cwd = os.path.dirname(os.path.dirname(self.path))
        if not cwd:
            cwd = "."
        self._call(params, self.err_branch, cwd)

    def is_modified(self):
        # To replace with bzr cli, we need to be able to detect
        # changes to a wc @ a rev or @ trunk.
        tree = WorkingTree.open(self.path)
        return tree.has_changes()


class Git(Vcs):

    def get_cur_rev(self):
        params = ["git", "rev-parse", "HEAD"]
        return self._call(params, self.err_cur_rev)

    def update(self, rev=None):
        params = ["git", "reset", "--merge"]
        if rev:
            params.append(rev)
        self._call(params, self.err_update)

    def branch(self):
        params = ["git", "clone", "--depth", "1", self.branch]
        self._call(params, self.err_branch, os.path.dirname(self.path))

    def is_modified(self):
        params = ["git", "stat", "-s"]
        return bool(self._call(params, self.err_is_mod).strip())

    def get_origin(self):
        params = ["git", "config", "--get", "remote.origin.url"]
        return self._call(params, "")



import logging
import time

from .base import BaseAction
from ..env import watchers
from ..errors import UnitErrors
from ..utils import ErrorExit


class Importer(BaseAction):

    log = logging.getLogger("deployer.import")

    def __init__(self, env, deployment, options):
        self.options = options
        self.env = env
        self.deployment = deployment

    def add_units(self):
        self.log.debug("Adding units...")
        # Add units to existing services that don't match count.
        env_status = self.env.status()
        reloaded = False

        for svc in self.deployment.get_services():
            cur_units = len(env_status['services'][svc.name].get('units', ()))
            delta = (svc.num_units - cur_units)

            if delta <= 0:
                self.log.debug(
                    " Service %r does not need any more units added.",
                    svc.name)
                continue

            charm = self.deployment.get_charm_for(svc.name)
            if charm.is_subordinate():
                self.log.warning(
                    "Config specifies num units for subordinate: %s",
                    svc.name)
                continue

            self.log.info(
                "Adding %d more units to %s" % (abs(delta), svc.name))
            if svc.unit_placement:
                # Reload status once after non placed services units are done.
                if reloaded is False:
                    # Crappy workaround juju-core api inconsistency
                    time.sleep(5.1)
                    env_status = self.env.status()
                    reloaded = True

                placement = self.deployment.get_unit_placement(svc, env_status)
                for mid in range(cur_units, svc.num_units):
                    self.env.add_unit(svc.name, placement.get(mid))
            else:
                self.env.add_units(svc.name, abs(delta))

    def get_charms(self):
        # Get Charms
        self.log.debug("Getting charms...")
        self.deployment.fetch_charms(
            update=self.options.update_charms,
            no_local_mods=self.options.no_local_mods)

        # Load config overrides/includes and verify rels after we can
        # validate them.
        self.deployment.resolve(self.options.overrides or ())

    def deploy_services(self):
        self.log.info("Deploying services...")
        self.log.debug(self.env)
        env_status = self.env.status()
        reloaded = False

        for svc in self.deployment.get_services():
            if svc.name in env_status['services']:
                self.log.debug(
                    " Service %r already deployed. Skipping" % svc.name)
                continue

            charm = self.deployment.get_charm_for(svc.name)
            self.log.info(
                " Deploying service %s using %s", svc.name, charm.charm_url)

            if svc.unit_placement:
                # We sorted all the non placed services first, so we only
                # need to update status once after we're done with them.
                if not reloaded:
                    self.log.debug(
                        " Refetching status for placement deploys")
                    time.sleep(5.1)
                    env_status = self.env.status()
                    reloaded = True
                num_units = 1
            else:
                num_units = svc.num_units

            placement = self.deployment.get_unit_placement(svc, env_status)

            if charm.is_subordinate():
                num_units = None

            self.env.deploy(
                svc.name,
                charm.charm_url,
                self.deployment.repo_path,
                svc.config,
                svc.constraints,
                num_units,
                placement.get(0))

            if svc.annotations:
                self.log.debug(" Setting annotations")
                self.env.set_annotation(svc.name, svc.annotations)

            if self.options.deploy_delay:
                self.log.debug(" Waiting for deploy delay")
                time.sleep(self.options.deploy_delay)

    def add_relations(self):
        self.log.info("Adding relations...")

        # Relations
        status = self.env.status()
        created = False

        for end_a, end_b in self.deployment.get_relations():
            if self._rel_exists(status, end_a, end_b):
                continue
            self.log.info(" Adding relation %s <-> %s", end_a, end_b)
            self.env.add_relation(end_a, end_b)
            created = True
            # per the original, not sure the use case.
            #self.log.debug(" Waiting 5s before next relation")
            #time.sleep(5)
        return created

    def _rel_exists(self, status, end_a, end_b):
        # Checks for a named relation on one side that matches the local
        # endpoint and remote service.
        (name_a, name_b, rem_a, rem_b) = (end_a, end_b, None, None)

        if ":" in end_a:
            name_a, rem_a = end_a.split(":", 1)
        if ":" in end_b:
            name_b, rem_b = end_b.split(":", 1)

        rels_svc_a = status['services'][name_a].get('relations', {})

        found = False
        for r, related in rels_svc_a.items():
            if name_b in related:
                if rem_a and not r in rem_a:
                    continue
                found = True
                break
        if found:
            return True
        return False

    def check_timeout(self):
        timeout = self.options.timeout - (time.time() - self.start_time)
        if timeout < 0:
            self.log.error("Reached deployment timeout.. exiting")
            raise ErrorExit()
        return timeout

    def wait_for_units(self, ignore_errors=False):
        timeout = self.check_timeout()
        # Set up the callback to be called in case of unit errors: if
        # ignore_errors is True errors are just logged, otherwise we exit the
        # program.
        if ignore_errors:
            on_errors = watchers.log_on_errors(self.env)
        else:
            on_errors = watchers.exit_on_errors(self.env)
        self.env.wait_for_units(
            int(timeout), watch=self.options.watch,
            services=self.deployment.get_service_names(), on_errors=on_errors)

    def run(self):
        options = self.options
        self.start_time = time.time()

        # Get charms
        self.get_charms()
        if options.branch_only:
            return

        if options.bootstrap:
            self.env.bootstrap()
        self.env.connect()

        self.deploy_services()

        # Workaround api issue in juju-core, where any action takes 5s
        # to be consistent to subsequent watch api interactions, see
        # http://pad.lv/1203105 which will obviate this wait.
        time.sleep(5.1)
        self.add_units()

        ignore_errors = bool(options.retry_count) or options.ignore_errors
        self.log.debug("Waiting for units before adding relations")
        self.wait_for_units(ignore_errors=ignore_errors)

        # Reset our environment connection, as it may grow stale during
        # the watch (we're using a sync client so not responding to pings
        # unless actively using the conn).
        self.env.close()
        self.env.connect()

        self.check_timeout()
        rels_created = self.add_relations()

        # Wait for the units to be up before waiting for rel stability.
        if rels_created:
            self.log.debug(
                "Waiting for relation convergence %ds", options.rel_wait)
            time.sleep(options.rel_wait)
            self.wait_for_units(ignore_errors=ignore_errors)

        if options.retry_count:
            self.log.info("Looking for errors to auto-retry")
            self.env.resolve_errors(
                options.retry_count,
                options.timeout - time.time() - self.start_time)

        # Finally expose things
        for svc in self.deployment.get_services():
            if svc.expose:
                self.log.info(" Exposing service %r" % svc.name)
                self.env.expose(svc.name)



import logging

from ..utils import (
    _get_juju_home, path_join, yaml_load, ErrorExit,
    yaml_dump, temp_file, _check_call)


class BaseEnvironment(object):

    log = logging.getLogger("deployer.env")

    @property
    def env_config_path(self):
        jhome = _get_juju_home()
        env_config_path = path_join(jhome, 'environments.yaml')
        return env_config_path

    def _check_call(self, *args, **kwargs):
        if self.options and self.options.retry_count:
            kwargs['max_retry'] = self.options.retry_count
        return _check_call(*args, **kwargs)

    def _named_env(self, params):
        if self.name:
            params.extend(["-e", self.name])
        return params

    def _get_env_config(self):
        with open(self.env_config_path) as fh:
            config = yaml_load(fh.read())
            if self.name:
                if self.name not in config['environments']:
                    self.log.error("Environment %r not found", self.name)
                    raise ErrorExit()
                return config['environments'][self.name]
            else:
                env_name = config.get('default')
                if env_name is None:
                    if len(config['environments'].keys()) == 1:
                        env_name = config['environments'].keys().pop()
                    else:
                        self.log.error("Ambigious operation environment")
                        raise ErrorExit()
                return config['environments'][env_name]

    def _write_config(self, svc_name, config, fh):
        fh.write(yaml_dump({svc_name: config}))
        fh.flush()

#    def _write_config(self, svc_name, config, fh):
#        fh.write(yaml_dump(config))
#        fh.flush()

    def _get_units_in_error(self, status=None):
        units = []
        if status is None:
            status = self.status()
        for s in status.get('services', {}).keys():
            for uid, u in status['services'][s].get('units', {}).items():
                if 'error' in u['agent-state']:
                    units.append(uid)
                for uid, u in u.get('subordinates', {}).items():
                    if 'error' in u['agent-state']:
                        units.append(uid)
        return units

    def bootstrap(self):
        self.log.info("bootstraping, this might take a while...")
        params = self._named_env(["juju", "bootstrap"])
        self._check_call(
            params, self.log, "Failed to bootstrap")

        # block until topology is returned
        self.get_cli_status()
        self.log.info(" Bootstrap complete")

    def deploy(self, name, charm_url,
               repo=None, config=None,
               constraints=None, num_units=1, force_machine=None):
        params = self._named_env(["juju", "deploy"])
        with temp_file() as fh:
            if config:
                fh.write(yaml_dump({name: config}))
                fh.flush()
                params.extend(["--config", fh.name])
            if constraints:
                params.extend(['--constraints', constraints])
            if num_units not in (1, None):
                params.extend(["--num-units", str(num_units)])
            if charm_url.startswith('local'):
                if repo == "":
                    repo = "."
                params.extend(["--repository=%s" % repo])
            if force_machine is not None:
                params.extend(["--to=%s" % force_machine])

            params.extend([charm_url, name])
            self._check_call(
                params, self.log, "Error deploying service %r", name)

    def expose(self, name):
        params = self._named_env(["juju", "expose", name])
        self._check_call(
            params, self.log, "Error exposing service %r", name)

    def terminate_machine(self, mid, wait=False):
        """Terminate a machine.

        The machine can't have any running units, after removing the units or
        destroying the service, use wait_for_units to know when its safe to
        delete the machine (ie units have finished executing stop hooks and are
        removed)
        """
        if ((isinstance(mid, int) and mid == 0) or
           (mid.isdigit() and int(mid) == 0)):
            # Don't kill state server
            raise RuntimeError("Can't terminate machine 0")
        params = self._named_env(["juju", "terminate-machine"])
        params.append(mid)
        try:
            self._check_call(
                params, self.log, "Error terminating machine %r" % mid)
        except ErrorExit, e:
            if ("machine %s does not exist" % mid) in e.error.output:
                return
            raise

    def get_service_address(self, svc_name):
        status = self.get_cli_status()
        if svc_name not in status['services']:
            self.log.warning("Service %s does not exist", svc_name)
            return None
        svc = status['services'][svc_name]
        if 'subordinate-to' in svc:
            ps = svc['subordinate-to'][0]
            self.log.info(
                'Service %s is a subordinate to %s, finding principle service'
                % (svc_name, ps))
            return self.get_service_address(svc['subordinate-to'][0])

        units = svc.get('units', {})
        unit_keys = list(sorted(units.keys()))
        if unit_keys:
            return units[unit_keys[0]].get('public-address', '')
        self.log.warning("Service %s has no units" % svc_name)

    def get_cli_status(self):
        params = self._named_env(["juju", "status"])
        with open('/dev/null', 'w') as fh:
            output = self._check_call(
                params, self.log, "Error getting status, is it bootstrapped?",
                stderr=fh)
        status = yaml_load(output)
        return status

    def add_unit(self, service_name, machine_spec):
        raise NotImplementedError()

    def set_annotation(self, entity, annotations):
        raise NotImplementedError()



import logging
import mock
import os
import sys
import unittest

from deployer.config import ConfigStack
from deployer.action.importer import Importer
from deployer.utils import yaml_dump, yaml_load

from base import Base, TEST_OFFLINE


class Options(dict):

    def __getattr__(self, key):
        return self[key]


class ImporterTest(Base):

    def setUp(self):
        self.juju_home = self.mkdir()
        self.change_environment(JUJU_HOME=self.juju_home)
        self.options = Options({
            'bootstrap': False,
            'branch_only': False,
            'configs': [os.path.join(self.test_data_dir, 'wiki.yaml')],
            'debug': True,
            'deploy_delay': 0,
            'destroy_services': None,
            'diff': False,
            'find_service': None,
            'ignore_errors': False,
            'list_deploys': False,
            'no_local_mods': True,
            'overrides': None,
            'rel_wait': 60,
            'retry_count': 0,
            'series': None,
            'terminate_machines': False,
            'timeout': 2700,
            'update_charms': False,
            'verbose': True,
            'watch': False})

    @unittest.skipIf(TEST_OFFLINE,
                     "Requires configured bzr launchpad id and network access")
    @mock.patch('deployer.action.importer.time')
    def test_importer(self, mock_time):
        # Trying to track down where this comes from http://pad.lv/1243827
        stack = ConfigStack(self.options.configs)
        deploy = stack.get('wiki')
        env = mock.MagicMock()
        importer = Importer(env, deploy, self.options)
        importer.run()

        config = {'name': '$hi_world _are_you_there? {guess_who}'}
        self.assertEqual(
            env.method_calls[3], mock.call.deploy(
                'wiki', 'cs:precise/mediawiki', '', config, None, 1, None))

        self.assertEqual(
            yaml_load(yaml_dump(yaml_load(yaml_dump(config)))),
            config)



"""A collection of juju-core environment watchers."""

from jujuclient import WatchWrapper

from ..utils import ErrorExit


class WaitForMachineTermination(WatchWrapper):
    """Wait until the given machines are terminated."""

    def __init__(self, watch, machines):
        super(WaitForMachineTermination, self).__init__(watch)
        self.machines = set(machines)
        self.known = set()

    def process(self, entity_type, change, data):
        if entity_type != 'machine':
            return
        if change == 'remove' and data['Id'] in self.machines:
            self.machines.remove(data['Id'])
        else:
            self.known.add(data['Id'])

    def complete(self):
        for m in self.machines:
            if m in self.known:
                return False
        return True


class WaitForUnits(WatchWrapper):
    """Wait for units of the environment to reach a particular goal state.

    If services are provided, only consider the units belonging to the given
    services.
    If the on_errors callable is provided, call the given function each time a
    change set is processed and a new unit is found in an error state. The
    callable is called passing a list of units' data corresponding to the units
    in an error state.
    """
    def __init__(
            self, watch, goal_state='started', services=None, on_errors=None):
        super(WaitForUnits, self).__init__(watch)
        self.goal_state = goal_state
        self.services = services
        self.on_errors = on_errors
        # The units dict maps unit names to units data.
        self.units = {}
        # The units_in_error list contains the names of the units in error.
        self.units_in_error = []

    def process(self, entity, action, data):
        if entity != 'unit':
            return
        if (self.services is None) or (data['Service'] in self.services):
            unit_name = data['Name']
            if action == 'remove' and unit_name in self.units:
                del self.units[unit_name]
            else:
                self.units[unit_name] = data

    def complete(self):
        ready = True
        new_errors = []
        goal_state = self.goal_state
        on_errors = self.on_errors
        units_in_error = self.units_in_error
        for unit_name, data in self.units.items():
            status = data['Status']
            if status == 'error':
                if unit_name not in units_in_error:
                    units_in_error.append(unit_name)
                    new_errors.append(data)
            elif status != goal_state:
                ready = False
        if new_errors and goal_state != 'removed' and callable(on_errors):
            on_errors(new_errors)
        return ready


def log_on_errors(env):
    """Return a function receiving errors and logging them.

    The resulting function is suitable to be used as the on_errors callback
    for WaitForUnits (see above).
    """
    return env.log_errors


def exit_on_errors(env):
    """Return a function receiving errors, logging them and exiting the app.

    The resulting function is suitable to be used as the on_errors callback
    for WaitForUnits (see above).
    """
    def callback(errors):
        log_on_errors(env)(errors)
        raise ErrorExit()
    return callback


def raise_on_errors(exception):
    """Return a function receiving errors and raising the given exception.

    The resulting function is suitable to be used as the on_errors callback
    for WaitForUnits (see above).
    """
    def callback(errors):
        raise exception(errors)
    return callback



import logging
import os
import tempfile
import yaml

from deployer.deployment import Deployment
from deployer.config import ConfigStack
from deployer.utils import ErrorExit

from .base import Base


class ConfigTest(Base):

    def setUp(self):
        self.output = self.capture_logging(
            "deployer.config", level=logging.DEBUG)

    def test_config_basic(self):
        config = ConfigStack(['configs/ostack-testing-sample.cfg'])
        config.load()
        self.assertEqual(
            config.keys(),
            [u'openstack-precise-ec2',
             u'openstack-precise-ec2-trunk',
             u'openstack-ubuntu-testing'])
        self.assertRaises(ErrorExit, config.get, 'zeeland')
        result = config.get("openstack-precise-ec2")
        self.assertTrue(isinstance(result, Deployment))

    def test_config(self):
        config = ConfigStack([
            os.path.join(self.test_data_dir, "stack-default.cfg"),
            os.path.join(self.test_data_dir, "stack-inherits.cfg")])
        config.load()
        self.assertEqual(
            config.keys(),
            [u'my-files-frontend-dev', u'wordpress'])
        deployment = config.get("wordpress")
        self.assertTrue(deployment)

    def test_config_include_file(self):
        config = ConfigStack([
            os.path.join(self.test_data_dir, "stack-includes.cfg")])
        config.load()
        # ensure picked up stacks from both files
        self.assertEqual(
            config.keys(),
            [u'my-files-frontend-dev', u'wordpress'])

        # ensure inheritance was adhered to during cross-file load
        wordpress = [s.name for s in config.get('wordpress').get_services()]
        my_app = [s.name for s in
                  config.get('my-files-frontend-dev').get_services()]
        self.assertTrue(set(wordpress).issubset(set(my_app)))

    def test_inherits_config_overridden(self):
        config = ConfigStack([
            os.path.join(self.test_data_dir, "stack-default.cfg"),
            os.path.join(self.test_data_dir, "stack-inherits.cfg")])
        config.load()
        deployment = config.get('my-files-frontend-dev')
        db = deployment.get_service('db')
        # base deployment (wordpress)'s db tuning level should have been
        # over-ridden
        self.assertEquals(db.config.get('tuning-level'), 'fastest')

    def test_multi_inheritance_multi_files(self):
        config = ConfigStack([
            os.path.join(self.test_data_dir, "openstack", "openstack.cfg"),
            os.path.join(self.test_data_dir, "openstack", "ubuntu_base.cfg"),
            os.path.join(
                self.test_data_dir, "openstack", "openstack_base.cfg"),
        ])
        self._test_multiple_inheritance(config)

    def test_multi_inheritance_multi_included_files(self):
        # openstack.cfg
        #  includes -> ubuntu_base.cfg includes
        #               includes -> openstack_base.cfg
        test_conf = yaml.load(open(
            os.path.join(self.test_data_dir, "openstack", "openstack.cfg")))
        includes = [
            os.path.join(self.test_data_dir, "openstack", "ubuntu_base.cfg"),
            os.path.join(self.test_data_dir, "openstack", "openstack_base.cfg")
        ]
        for key in ['include-config', 'include-configs']:
            test_conf[key] = includes
            with tempfile.NamedTemporaryFile() as tmp_cfg:
                tmp_cfg.write(yaml.dump(test_conf))
                tmp_cfg.flush()
                config = ConfigStack([tmp_cfg.name])
                self._test_multiple_inheritance(config)
            del test_conf[key]

    def test_multi_inheritance_included_multi_configs(self):
        # openstack.cfg
        #  includes -> [ubuntu_base.cfg, openstack_base.cfg]
        config = ConfigStack([
            os.path.join(self.test_data_dir, "openstack", "openstack.cfg"),
        ])
        self._test_multiple_inheritance(config)

    def _test_multiple_inheritance(self, config):
        config.load()

        deployment = config.get('precise-grizzly')
        services = [s.name for s in list(deployment.get_services())]
        self.assertEquals(['mysql', 'nova-cloud-controller'], services)

        nova = deployment.get_service('nova-cloud-controller')
        self.assertEquals(nova.config['openstack-origin'],
                          'cloud:precise-grizzly')

        deployment = config.get('precise-grizzly-quantum')
        services = [s.name for s in list(deployment.get_services())]
        self.assertEquals(services,
                          ['mysql', 'nova-cloud-controller',
                           'quantum-gateway'])
        nova = deployment.get_service('nova-cloud-controller')
        self.assertEquals(nova.config['network-manager'], 'Quantum')
        self.assertEquals(nova.config['openstack-origin'],
                          'cloud:precise-grizzly')
        ex_rels = [('quantum-gateway', 'nova-cloud-controller'),
                   ('quantum-gateway', 'mysql'),
                   ('nova-cloud-controller', 'mysql')]
        self.assertEquals(ex_rels, list(deployment.get_relations()))

    def test_config_series_override(self):
        config = ConfigStack(['configs/wiki.yaml'], 'trusty')
        config.load()
        result = config.get("wiki")
        self.assertTrue(isinstance(result, Deployment))
        self.assertEquals(result.series, 'trusty')


class NetworkConfigFetchingTests(Base):
    """Configuration files can be specified via URL that is then fetched."""

    def setUp(self):
        self.output = self.capture_logging(
            "deployer.config", level=logging.DEBUG)

    def test_urls_are_fetched(self):
        # If a config file is specified as a URL, that URL is fetched and
        # placed at a temporary location where it is read and treated as a
        # regular config file.
        CONFIG_URL = 'http://site.invalid/config-1'
        config = ConfigStack([])
        config.config_files = [CONFIG_URL]

        class FauxResponse(file):
            def getcode(self):
                return 200

        def faux_urlopen(url):
            self.assertEqual(url, CONFIG_URL)
            return FauxResponse('configs/ostack-testing-sample.cfg')

        config.urlopen = faux_urlopen
        config.load()
        self.assertEqual(
            config.keys(),
            [u'openstack-precise-ec2',
             u'openstack-precise-ec2-trunk',
             u'openstack-ubuntu-testing'])
        self.assertRaises(ErrorExit, config.get, 'zeeland')
        result = config.get("openstack-precise-ec2")
        self.assertTrue(isinstance(result, Deployment))

    def test_unfetchable_urls_generate_an_error(self):
        # If a config file is specified as a URL, that URL is fetched and
        # placed at a temporary location where it is read and treated as a
        # regular config file.
        CONFIG_URL = 'http://site.invalid/config-1'
        config = ConfigStack([])
        config.config_files = [CONFIG_URL]

        class FauxResponse(file):
            def getcode(self):
                return 400

        def faux_urlopen(url):
            self.assertEqual(url, CONFIG_URL)
            return FauxResponse('configs/ostack-testing-sample.cfg')
        config.urlopen = faux_urlopen
        self.assertRaises(ErrorExit, config.load)



"""Tests for the GUIEnvironment."""

import unittest
import mock

from deployer.env.gui import (
    GUIEnvironment,
)


@mock.patch('deployer.env.go.EnvironmentClient')
class TestGUIEnvironment(unittest.TestCase):

    endpoint = 'wss://api.example.com:17070'
    password = 'Secret!'

    def setUp(self):
        self.env = GUIEnvironment(self.endpoint, self.password)

    def test_connect(self, mock_client):
        # The environment uses the provided endpoint and password to connect
        # to the Juju API server.
        self.env.connect()
        mock_client.assert_called_once_with(self.endpoint)
        mock_client().login.assert_called_once_with(self.password)

    def test_multiple_connections(self, mock_client):
        # The environment does not attempt a second connection if it is already
        # connected to the API backend.
        self.env.connect()
        self.env.connect()
        self.assertEqual(1, mock_client.call_count)

    def test_close(self, mock_client):
        # The client attribute is set to None when the connection is closed.
        self.env.connect()
        self.env.close()
        self.assertIsNone(self.env.client)

    def test_deploy(self, mock_client):
        # The environment uses the API to deploy charms.
        # Constraints are converted to numeric values before calling the
        # client deploy.
        self.env.connect()
        config = {'foo': 'bar'}
        constraints = {'cpu-cores': '4', 'mem': '5G'}
        # Deploy a service: the repo argument is ignored.
        self.env.deploy(
            'myservice', 'cs:precise/service-42', repo='/my/repo/',
            config=config, constraints=constraints, num_units=2,
            force_machine=1)
        expected = {'cpu-cores': 4, 'mem': 5 * 1024}
        mock_client().deploy.assert_called_once_with(
            'myservice', 'cs:precise/service-42', config=config,
            constraints=expected, num_units=2, machine_spec=1)

    def test_deploy_unqualified_url(self, mock_client):
        # Show that an unqualified URL is given the latest revision.
        self.env.connect()

        def mocked_get_qualified_charm_url(url):
            return url + '-98'

        path = 'deployer.env.gui.get_qualified_charm_url'
        with mock.patch(path, mocked_get_qualified_charm_url):
            self.env.deploy(
                'myservice', 'cs:precise/service', repo='/my/repo/')
        mock_client().deploy.assert_called_once_with(
            'myservice', 'cs:precise/service-98', config=None,
            constraints=None, num_units=1, machine_spec=None)



from setuptools import setup, find_packages


setup(
    name="juju-deployer",
    version="0.3.9",
    description="A tool for deploying complex stacks with juju.",
    long_description=open("README").read(),
    author="Kapil Thangavelu",
    author_email="kapil.foss@gmail.com",
    url="http://launchpad.net/juju-deployer",
    install_requires=["jujuclient>=0.15", "PyYAML>=3.10"],
    packages=find_packages(),
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python",
        "Topic :: Internet",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Intended Audience :: System Administrators",
        "Intended Audience :: Developers"],
    test_suite="deployer.tests",
    entry_points={
        "console_scripts": [
            'juju-deployer = deployer.cli:main']})
